# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """


from .iterator import IteratorSource

class GeneratorSource(IteratorSource):
    pass